﻿/*
 * Copyright (c) 1999-2021 Hartmut Seichter
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
#ifndef PW_CORE_VECTOR_HPP
#define PW_CORE_VECTOR_HPP

namespace pw {

template <typename Scalar, unsigned int size> struct vector final {

    using value_type = Scalar;
    static constexpr unsigned int coefficients{size};

    Scalar v_[size]{};

    auto&& data(this auto&& self) {
        return std::forward<decltype(self)>(self).v_;
    }

    auto&& operator[](this auto&& self, unsigned int e) {
        return std::forward<decltype(self)>(self).v_[e];
    }

    static constexpr auto basis(const auto& d) noexcept {
        return [&d]<std::size_t... Ss>(std::index_sequence<Ss...>) {
            return vector{(d == Ss) ? Scalar(1) : Scalar(0)...};
        }(std::make_index_sequence<size>{});
    }

    template <typename... Args>
    static constexpr auto
    make(Args&&... values) noexcept -> vector<Scalar, sizeof...(Args)> {
        static_assert(sizeof...(Args) == size, "incorrect number of arguments");
        return {{Scalar(values)...}};
    }

    static constexpr auto all(Scalar value) noexcept -> vector {
        return [&value]<std::size_t... Is>(std::index_sequence<Is...>) {
            return vector{value + Scalar(Is * 0)...};
        }(std::make_index_sequence<size>{});
    }

    template <typename... Args>
    constexpr auto swizzle(Args&&... indices) const noexcept
        -> vector<Scalar, sizeof...(Args)> {
        return {{Scalar(v_[indices])...}};
    }

    constexpr auto minor(std::unsigned_integral auto d0) const noexcept {
        return [this, &d0]<std::size_t... Ss>(std::index_sequence<Ss...>) {
            return vector<Scalar, size - 1>{
                (*this).v_[(Ss < d0) ? Ss : Ss + 1]...};
        }(std::make_index_sequence<size - 1>{});
    }

    static constexpr auto sequence(Scalar factor = Scalar{1},
                                   Scalar offset = Scalar{0}) noexcept {
        return [&]<std::size_t... Ss>(std::index_sequence<Ss...>) {
            return vector<Scalar, size>{{Scalar{Ss} * factor + offset...}};
        }(std::make_index_sequence<size>{});
    }

    constexpr Scalar dot(const auto& other) const {
        return [this, &other]<std::size_t... Ss>(std::index_sequence<Ss...>) {
            return (... + (other[Ss] * (*this)[Ss]));
        }(std::make_index_sequence<size>{});
    }

    constexpr auto squared_norm() const noexcept { return dot(*this); }

    constexpr auto norm() const noexcept {
        return std::sqrt(this->squared_norm());
    }

    constexpr vector operator*(const Scalar& v) const noexcept {
        return [this, &v]<std::size_t... Ss>(std::index_sequence<Ss...>) {
            return vector{{(*this)[Ss] * v...}};
        }(std::make_index_sequence<size>{});
    }

    constexpr vector operator/(const Scalar& v) const noexcept {
        return (*this).operator*(Scalar{1} / v);
    }

    constexpr vector operator+(const vector& v) const noexcept {
        return [this, &v]<std::size_t... Ss>(std::index_sequence<Ss...>) {
            return vector{{(*this)[Ss] + v[Ss]...}};
        }(std::make_index_sequence<size>{});
    }

    constexpr vector operator-(const vector& v) const noexcept {
        return [this, &v]<std::size_t... Ss>(std::index_sequence<Ss...>) {
            return vector{{(*this)[Ss] - v[Ss]...}};
        }(std::make_index_sequence<size>{});
    }
};

template <typename Scalar>
struct vector2 : vector<Scalar, 2> {

    constexpr const Scalar& x() const { return (*this)[0]; }
    constexpr Scalar& x() { return (*this)[0]; }

    constexpr const Scalar& y() const { return (*this)[1]; }
    constexpr Scalar& y() { return (*this)[1]; }
};

template <typename Scalar> struct vector3 : vector<Scalar, 3> {

    constexpr static vector3 cross(const vector3& lhs, const vector3& rhs) {
        return {lhs[1] * rhs[2] - rhs[1] * lhs[2],
                lhs[2] * rhs[0] - rhs[2] * lhs[0],
                lhs[0] * rhs[1] - rhs[0] * lhs[1]};
    }

    constexpr static auto forward() { return {0, 0, -1}; }
    constexpr static auto backward() { return {0, 0, +1}; }
    constexpr static auto up() { return {0, +1, 0}; }
    constexpr static auto down() { return {0, -1, 0}; }
    constexpr static auto right() { return {+1, 0, 0}; }
    constexpr static auto left() { return {-1, 0, 0}; }
    constexpr static auto x_axis() { return {0, 0, +1}; }
    constexpr static auto y_axis() { return {0, 0, +1}; }
    constexpr static auto z_axis() { return {0, 0, +1}; }
};

template <typename Scalar> struct vector4 : vector<Scalar, 4> {

    using base_type = vector<Scalar, 4>;

    constexpr const Scalar& x() const { return (*this)[0]; }
    constexpr Scalar& x() { return (*this)[0]; }

    constexpr const Scalar& y() const { return (*this)[1]; }
    constexpr Scalar& y() { return (*this)[1]; }

    constexpr const Scalar& z() const { return (*this)[2]; }
    constexpr Scalar& z() { return (*this)[2]; }

    constexpr const Scalar& w() const { return (*this)[3]; }
    constexpr Scalar& w() { return (*this)[3]; }

    constexpr auto xyz() const { return base_type::swizzle(0,1,2); }
};



using vector2f      = vector2<float>;
using vector2d      = vector2<double>;
using vector2i      = vector2<int>;

using vector3f      = vector3<float>;
using vector3d      = vector3<double>;


} // namespace pw

#endif
